from apifairy import APIFairy
from config import Config
from docker import DockerClient
from flask import Flask
from flask_apscheduler import APScheduler
from flask_assets import Environment
from flask_breadcrumbs import Breadcrumbs, default_breadcrumb_root
from flask_login import LoginManager
from flask_mail import Mail
from flask_marshmallow import Marshmallow
from flask_migrate import Migrate
from flask_paranoid import Paranoid
from flask_socketio import SocketIO
from flask_sqlalchemy import SQLAlchemy
from flask_hashids import Hashids
from werkzeug.exceptions import HTTPException


apifairy = APIFairy()
assets = Environment()
breadcrumbs = Breadcrumbs()
db = SQLAlchemy()
docker_client = DockerClient()
hashids = Hashids()
login = LoginManager()
login.login_view = 'auth.login'
login.login_message = 'Please log in to access this page.'
ma = Marshmallow()
mail = Mail()
migrate = Migrate(compare_type=True)
paranoid = Paranoid()
paranoid.redirect_view = '/'
scheduler = APScheduler()
socketio = SocketIO()


def create_app(config: Config = Config) -> Flask:
    ''' Creates an initialized Flask (WSGI Application) object. '''
    app = Flask(__name__)
    app.config.from_object(config)
    config.init_app(app)
    docker_client.login(
        app.config['NOPAQUE_DOCKER_REGISTRY_USERNAME'],
        password=app.config['NOPAQUE_DOCKER_REGISTRY_PASSWORD'],
        registry=app.config['NOPAQUE_DOCKER_REGISTRY']
    )

    apifairy.init_app(app)
    assets.init_app(app)
    breadcrumbs.init_app(app)
    db.init_app(app)
    hashids.init_app(app)
    login.init_app(app)
    ma.init_app(app)
    mail.init_app(app)
    migrate.init_app(app, db)
    paranoid.init_app(app)
    scheduler.init_app(app)
    socketio.init_app(app, message_queue=app.config['NOPAQUE_SOCKETIO_MESSAGE_QUEUE_URI'])  # noqa

    from .admin import bp as admin_blueprint
    default_breadcrumb_root(admin_blueprint, '.admin')
    app.register_blueprint(admin_blueprint, url_prefix='/admin')

    from .api import bp as api_blueprint
    app.register_blueprint(api_blueprint, url_prefix='/api')

    from .auth import bp as auth_blueprint
    default_breadcrumb_root(auth_blueprint, '.')
    app.register_blueprint(auth_blueprint)

    from .contributions import bp as contributions_blueprint
    default_breadcrumb_root(contributions_blueprint, '.contributions')
    app.register_blueprint(contributions_blueprint, url_prefix='/contributions')

    from .corpora import bp as corpora_blueprint
    from .corpora.cqi_over_sio import CQiNamespace
    default_breadcrumb_root(corpora_blueprint, '.corpora')
    app.register_blueprint(corpora_blueprint, cli_group='corpus', url_prefix='/corpora')
    socketio.on_namespace(CQiNamespace('/cqi_over_sio'))

    from .errors import bp as errors_bp
    app.register_blueprint(errors_bp)

    from .jobs import bp as jobs_blueprint
    default_breadcrumb_root(jobs_blueprint, '.jobs')
    app.register_blueprint(jobs_blueprint, url_prefix='/jobs')

    from .main import bp as main_blueprint
    default_breadcrumb_root(main_blueprint, '.')
    app.register_blueprint(main_blueprint, cli_group=None)

    from .services import bp as services_blueprint
    default_breadcrumb_root(services_blueprint, '.services')
    app.register_blueprint(services_blueprint, url_prefix='/services')

    from .settings import bp as settings_blueprint
    default_breadcrumb_root(settings_blueprint, '.settings')
    app.register_blueprint(settings_blueprint, url_prefix='/settings')

    from .users import bp as users_blueprint
    default_breadcrumb_root(users_blueprint, '.users')
    app.register_blueprint(users_blueprint, url_prefix='/users')

    return app
